# This is a very basic way of abstracting out the prompt structures to address different use cases
#  as well as special LLM model handling as well as make it easier to apply in the RAGAS test suite

# References
# AWS Bedrock
# https://docs.aws.amazon.com/bedrock/latest/userguide/general-guidelines-for-bedrock-users.html
# https://docs.aws.amazon.com/bedrock/latest/userguide/prompt-templates-and-examples.html
# Open AI
# https://platform.openai.com/docs/guides/prompt-engineering
# Anthropic
# https://docs.anthropic.com/claude/docs/prompt-engineering
# Google Gemini
# https://ai.google.dev/docs/prompt_best_practices



class BasePromptModel:
    prompt_model_name = "Generic Base Prompt Model"
    llm_temperature = 0.5
    llm_top_p = 0.5
    llm_index = "index-artemis-mission"

    # This is used to ask the question with retrieval context applied
    system_prompt_template = """You are an assistant for question-answering tasks. \
    Use the following pieces of retrieved context to answer the question. \
    If you don't know the answer, just say that you don't know. \
    Use three sentences maximum and keep the answer concise.\

    {context}
    
    {question} \n
    """

    # If there is conversational history, we use the LLM to summarize the question into a single question
    summary_prompt_template = """Given a chat history below and my latest question at the end
        which might reference context in the chat history, formulate a standalone question 
        which can be understood without the chat history. Do NOT answer the question, 
        just reformulate it if needed and otherwise return it as is.
    
        Chat history: \n
        {chat_history}
        
        \n\n
        Based on the above conversation my latest questions is this: \n
        {question}
        
        """


class ClaudDefaultPrompts(BasePromptModel):
    system_prompt_template = """
    You are an assistant for question-answering tasks..
    Answer the question asked which can be found at the end of this information.  
    If you don't know the answer, just say that you don't know.

    Answer the question based only on the following context -  <context>\n{context}</context>

    Please provide the one or more of the most relevant quotes from the context that support your answer inside the <proof> tags.

    My questions is this - 
    {question}

    Assistant:
    """

    summary_prompt_template = """Given a chat history and summary contained <conversational_history> below 
                    in the and the latest user question formulate a standalone question \
                    which can be understood without the chat history. Do NOT answer the question, \
                    just reformulate it if needed and otherwise return it as is.

                    Current conversation summary - \n<conversational_history>{chat_history}</conversational_history>

                    and here is my latest question - \n
                    {question}

                    \n\nAssistant:
                    """


class ClaudNasaSpokespersonPrompts(ClaudDefaultPrompts):
    system_prompt_template = """
    You are the spokesperson for NASA named Sally speaking to a  group of journalists about the Artemis space program.
    Answer the question asked which can be found at the end of this information.  keep the answer concise.
    Keep the tone friendly, positive, informative and engaging.
    If the question is not related to NASA or the Artemis program, just say "I am here to talk about NASA and our mission to the moon."
    Provide only the answer and don't follow the answer with a question to help more.

    Answer the question based only on the following context -  <context>\n{context}</context>

    Please provide the one or more of the most relevant quotes from the context that support your answer inside the <proof> tags.

    My questions is this - 
    {question}

    Assistant:
    """


class NasaSpokespersonPrompts(BasePromptModel):
    system_prompt_template = """
    You are the spokesperson for NASA speaking to a  \
    group of journalists.  Respond in a way that is informative and engaging. \
    Use the following pieces of retrieved context to answer the question. \
    keep the answer concise.
    
    {context}
      
    {question} \n
    """



from datetime import datetime
import boto3
from dotenv import find_dotenv, load_dotenv

from langchain.llms.bedrock import Bedrock
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnablePassthrough
from langchain_core.messages import AIMessage, HumanMessage
from langchain_openai import ChatOpenAI
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain.schema.runnable import RunnableConfig
from langchain.callbacks.base import BaseCallbackHandler

# Code in this project
from aws_rag_bot.aws_opensearch_vector_database import (
    EmbeddingTypes,
    OpenSearchVectorDBQuery
)
from aws_rag_bot.prompt_library import *


class LlmModelTypes:
    BEDROCK_LLAMA2 = "bedrock_llama2"
    BEDROCK_JURRASIC2_ULTRA = "bedrock_jurrasic2_ultra"
    BEDROCK_TITAN_EXPRESS = "bedrock_titan_express"
    BEDROCK_CLAUDE_INSTANT = "bedrock_claude_instant"
    BEDROCK_CLAUDE_21 = "bedrock_claude21"
    OPENAI_GPT4 = "openai_gpt4"
    OPENAI_GPT35 = "openai_gpt35"
    GOOGLE_GEMINI_PRO = "google_gemini_pro"


def structure_message_history(chat_history):
    # format of chat history is: {"question": "****", "response": "****"]
    if chat_history is None or len(chat_history) == 0:
        return []
    else:
        history_string = []
        for chat_tuple in chat_history:
            history_string.append(HumanMessage(content=chat_tuple["question"]))
            history_string.append(AIMessage(content=chat_tuple["response"]))
        return history_string


class RagChatbot:
    __llm_model = None
    __conversation_chain = None
    __vector_db = None
    __vector_db_embedding = None

    __chat_prompt = None
    __log_verbose = False
    __total_cost = 0
    __total_input_tokens = 0
    __total_output_tokens = 0
    __total_run_duration = 0
    __last_run_prompt = None
    __last_run_input_tokens = 0
    __last_run_output_tokens = 0
    __last_run_cost = 0
    __last_run_duration = 0

    __bedrock_model_def_llama2 = {
        "key": "bedrock_llama2",
        "name": "Bedrock Llama 2",
        "id": "meta.llama2-70b-chat-v1",
        "client_name": "bedrock-runtime",
        "region_name": "us-east-1",
        "kwargs": {
            "max_gen_len": 500,
            "temperature": 0.2,
            "top_p": 0.2,
        },
        "model_cost": {
            "input_token_cost": 0.00195 / 1000,
            "output_token_cost": 0.00256 / 1000
        }
    }

    __bedrock_model_def_jurrasic2_ultra = {
        "key": "bedrock_jurrasic2_ultra",
        "name": "Bedrock Jurrasic 2 Ultra",
        "id": "ai21.j2-ultra-v1",
        "client_name": "bedrock-runtime",
        "region_name": "us-east-1",
        "kwargs": {
            "maxTokens": 500,
            "temperature": 0.5,
            "topP": 0.5,
        },
        "model_cost": {
            "input_token_cost": 0.0188 / 1000,
            "output_token_cost": 0.0188 / 1000
        }
    }

    __bedrock_model_def_titan_express = {
        "key": "bedrock_titan_express",
        "name": "Bedrock Titan Express",
        "id": "amazon.titan-text-express-v1",
        "client_name": "bedrock-runtime",
        "region_name": "us-east-1",
        "kwargs": {
            "maxTokenCount": 1000,
            "temperature": 0.1,
            "topP": 0.5,
        },
        "model_cost": {
            "input_token_cost": 0.0008 / 1000,
            "output_token_cost": 0.0016 / 1000
        }
    }

    __bedrock_model_def_claude_instant = {
        "key": "bedrock_claude_instant",
        "name": "Bedrock Claude Instant",
        "id": "anthropic.claude-instant-v1",
        "client_name": "bedrock-runtime",
        "region_name": "us-east-1",
        "kwargs": {
            "max_tokens_to_sample": 1000,
            "temperature": 0.1,
            "top_p": 0.5,
            "top_k": 250,
        },
        "model_cost": {
            "input_token_cost": 0.00080 / 1000,
            "output_token_cost": 0.00240 / 1000
        }
    }

    __bedrock_model_def_claude21 = {
        "key": "bedrock_claude21",
        "name": "Bedrock Claude v2.1",
        "id": "anthropic.claude-v2:1",
        "client_name": "bedrock-runtime",
        "region_name": "us-east-1",
        "kwargs": {
            "max_tokens_to_sample": 1000,
            "temperature": 0.1,
            "top_p": 0.5,
            "top_k": 250,
        },
        "model_cost": {
            "input_token_cost": 0.00080 / 1000,
            "output_token_cost": 0.00240 / 1000
        }
    }

    __open_ai_model_def_gpt4 = {
        "key": "openai_gpt4",
        "name": "OpenAI GPT-4",
        "id": "gpt-4",
        "client_name": "openai",
        "kwargs": {
            "maxTokens": 500,
            "temperature": 0.5,
        },
        "model_cost": {
            "input_token_cost": 0.03 / 1000,
            "output_token_cost": 0.06 / 1000
        }
    }

    __open_ai_model_def_gpt35 = {
        "key": "openai_gpt35",
        "name": "OpenAI GPT-3.5 Turbo",
        "id": "gpt-3.5-turbo",
        "client_name": "openai",
        "kwargs": {
            "maxTokens": 500,
            "temperature": 0.5,
        },
        "model_cost": {
            "input_token_cost": 0.0005 / 1000,
            "output_token_cost": 0.0015 / 1000
        }
    }

    __google_gemini_pro = {
        "key": "google_gemini_pro",
        "name": "Google Gemini Pro",
        "id": "gemini-pro",
        "client_name": "google",
        "kwargs": {
            "maxTokens": 500,
            "temperature": 0.5,
        },
        "model_cost": {  # uses cost/char... estimated at 4 chars per token
            "input_token_cost": 0.000125 * 4 / 1000,
            "output_token_cost": 0.000375 * 4 / 1000
        }
    }

    __model_options = [
        __bedrock_model_def_llama2,
        __bedrock_model_def_jurrasic2_ultra,
        __bedrock_model_def_titan_express,
        __bedrock_model_def_claude_instant,
        __bedrock_model_def_claude21,
        __open_ai_model_def_gpt4,
        __open_ai_model_def_gpt35,
        __google_gemini_pro
    ]
    __current_model = None
    __prompt_model = None

    def token_count(self, string):
        return self.__llm_model.get_num_tokens(string)

    # Constructor
    def __init__(self, vector_db_domain
                 , model_key=LlmModelTypes.BEDROCK_TITAN_EXPRESS
                 , prompt_model=None
                 , embedding_model=EmbeddingTypes.BEDROCK_DEFAULT
                 , verbose=False):

        load_dotenv(find_dotenv())
        if prompt_model is None:
            self.__prompt_model = BasePromptModel()
        else:
            self.__prompt_model = prompt_model

        self.__log_verbose = verbose

        # Set the current model based on input
        for model_option in self.__model_options:
            if model_option['key'] == model_key:
                self.__current_model = model_option
                break

        # Set the index name based on the prompt model
        index_name = self.__prompt_model.llm_index

        # Initialize the LLM model, memory, and chain
        self.__llm_model = self.__get_llm_model()
        vector_db = OpenSearchVectorDBQuery(domain_name=vector_db_domain,
                                            index_name=index_name,
                                            embedding_model=embedding_model)
        self.__vector_db = vector_db.get_client()

    class RagCallback(BaseCallbackHandler):
        # https://how.wtf/how-to-count-amazon-bedrock-anthropic-tokens-with-langchain.html
        def __init__(self, llm, verbose=False):
            self.llm = llm
            self.prompt = None
            self.input_tokens = 0
            self.output_tokens = 0
            self.verbose = verbose

        def on_llm_start(self, input, prompts, **kwargs):
            self.prompt = prompts
            if self.verbose:
                print("--------------------- Callback On Start LLM Prompts --------------------------------------")
                print(prompts)
                print("------------------------------------------------------------------------------------------")
            for p in prompts:
                self.input_tokens = self.llm.get_num_tokens(p)

        def on_llm_end(self, output, **kwargs):
            results = output.flatten()
            if self.verbose:
                print("--------------------- Callback On End LLM Results --------------------------------------")
                print(results)
                print("----------------------------------------------------------------------------------------")
            for r in results:
                self.output_tokens = self.llm.get_num_tokens(r.generations[0][0].text)

        def cost(self, model_info):
            if 'model_cost' in model_info:
                return self.input_tokens * model_info['model_cost']['input_token_cost'] + self.output_tokens * \
                    model_info['model_cost']['output_token_cost']

        # def on_llm_new_token(self, token, **kwargs):
        #     Says I need streaming enabled for this to work. maybe later
        #     self.input_tokens += 1

    # Create the LLM model
    def __get_llm_model(self):
        if self.__log_verbose:
            print("RagChatbot: Initializing LLM model")
            print(f"Creating LLM model: {self.__current_model['name']}")
            print(f"  - client: {self.__current_model['client_name']}")
            print(f"  - id: {self.__current_model['id']}")

        llm_model_kwargs = self.__current_model['kwargs']

        # Override settings with the prompt model settings
        llm_model_kwargs['temperature'] = self.__prompt_model.llm_temperature
        if 'topP' in llm_model_kwargs:
            llm_model_kwargs['topP'] = self.__prompt_model.llm_top_p

        if 'top_p' in llm_model_kwargs:
            llm_model_kwargs['top_p'] = self.__prompt_model.llm_top_p

        llm_model = None
        # Handle creating Bedrock Models
        if self.__current_model['client_name'] == 'bedrock-runtime':
            # Explicitly create client with boto3 to get better control and transparency
            bedrock = boto3.client('bedrock-runtime', region_name=self.__current_model['region_name'])

            llm_model = Bedrock(
                model_id=self.__current_model['id'],
                client=bedrock,
                model_kwargs=llm_model_kwargs
            )

        elif self.__current_model['client_name'] == 'openai':
            temp = llm_model_kwargs['temperature']
            llm_model = ChatOpenAI(model=self.__current_model['id'], temperature=temp)

        elif self.__current_model['client_name'] == 'google':
            temp = llm_model_kwargs['temperature']
            llm_model = ChatGoogleGenerativeAI(model=self.__current_model['id']
                                               , temperature=temp
                                               , convert_system_message_to_human=True)


        else:
            raise Exception("Unsupported client_name: {}".format(self.__current_model['client_name']))

        return llm_model

    def get_llm_model(self):
        return self.__llm_model

    def ask_question(self, question, conversation_history=None, verbose=False):
        my_callback_handler = self.RagCallback(self.__llm_model, verbose=verbose)
        chain_config = RunnableConfig(callbacks=[my_callback_handler])

        # --- If we have history, then summarize along with the question to produce a new question ---
        # format of chat history is: [{"question": "****", "response": "****"},]
        #  and is managed and formulated by the client to this call
        restated_question = None
        if conversation_history is not None:
            summary_prompt = ChatPromptTemplate.from_template(self.__prompt_model.summary_prompt_template)
            history = ""
            for qa_pair in conversation_history:
                history += f"Question - {qa_pair['question']}\nResponse - {qa_pair['response']}\n"
            summary_request = summary_prompt.format(chat_history=history, question=question)
            restated_question = self.__llm_model.invoke(summary_request, config=chain_config)

        else:
            restated_question = question

        # --- Now we have a restated question, we ask the with the intended prompts ---
        # Get a retriever
        retriever = self.__vector_db.as_retriever()

        # Register my callback handler

        template = self.__prompt_model.system_prompt_template
        prompt = ChatPromptTemplate.from_template(template)

        # This takes the context from the retriever, based on the question, then
        #   passes the resulting context and question to the prompt template
        #   then executes the query and returns.
        my_chain = (
                {"context": retriever, "question": RunnablePassthrough()}
                | prompt
                | self.__llm_model
                | StrOutputParser()
        )
        # provide config to the chain, which at the moment includes the callback handler
        my_chain = my_chain.with_config(chain_config)
        start_time = datetime.now()
        if type(restated_question) == AIMessage:
            answer = my_chain.invoke(restated_question.content)
        else:
            answer = my_chain.invoke(restated_question)

        end_time = datetime.now()
        duration = end_time - start_time

        self.__last_run_input_tokens = my_callback_handler.input_tokens
        self.__last_run_output_tokens = my_callback_handler.output_tokens
        self.__last_run_cost = my_callback_handler.cost(self.__current_model)
        self.__total_cost += my_callback_handler.cost(self.__current_model)
        self.__total_input_tokens += my_callback_handler.input_tokens
        self.__total_output_tokens += my_callback_handler.output_tokens
        self.__last_run_prompt = my_callback_handler.prompt
        self.__last_run_duration = duration.total_seconds()
        self.__total_run_duration += duration.total_seconds()

        # if type(response) == AIMessage:
        #     return response
        # else:  # This is needed because some LLM's will return a string instead of a AIMessage, like Titan Express
        #     return AIMessage(content=response)

        if type(answer) == AIMessage:
            return answer.content
        else:
            return answer

    def get_model_run_summary(self):
        return {
            "total_cost": self.__total_cost,
            "total_input_tokens": self.__total_input_tokens,
            "total_output_tokens": self.__total_output_tokens,
            "total_run_duration": self.__total_run_duration,

            "last_run_input_tokens": self.__last_run_input_tokens,
            "last_run_output_tokens": self.__last_run_output_tokens,
            "last_run_cost": self.__last_run_cost,
            "last_run_duration": self.__last_run_duration,

            "model_name": self.__current_model,
            "vector_db_index": self.__prompt_model.llm_index
        }

    def get_last_run_prompt(self):
        return self.__last_run_prompt

    def get_vector_db(self):
        return self.__vector_db



import requests
from langchain_core.documents import Document
import boto3
import os
from dotenv import load_dotenv, find_dotenv


def get_all_shortcuts_as_documents(kustomer_function):
    # Get additional configuration
    load_dotenv(find_dotenv())
    kustomer_api_key = os.getenv('KUSTOMER_API_KEY')

    base_url = "https://api.kustomerapp.com"
    initial_url = f"{base_url}/{kustomer_function}"
    headers = {
        "Authorization": f"Bearer {kustomer_api_key}",
        "Content-Type": "application/json"
    }

    print(f"Fetching shortcuts from {initial_url}")
    documents = []
    url = initial_url
    while url:
        response = requests.get(url, headers=headers)

        if response.status_code == 200:
            data = response.json()
            for item in data.get('data', []):
                # Check if appDisabled or deleted are False
                if item['attributes'].get('appDisabled', True) or item['attributes'].get('deleted', True):
                    continue  # Skip this item

                name = item['attributes'].get('name', '')
                text = item['attributes']['draft'].get('text', '')
                page_content = f"{name}: {text}"
                metadata = {
                    "source": "Kustomer Shortcut",
                    "title": name,
                    "created_at": item['attributes'].get('updatedAt', '')
                }
                doc = Document(page_content=page_content, metadata=metadata)
                documents.append(doc)

            # Check if there's a next page
            next_link = data.get('links', {}).get('next')
            url = f"{base_url}{next_link}" if next_link else None
        else:
            raise Exception(f"Failed to fetch shortcuts. Status code: {response.status_code}")

    return documents



from langchain_community.document_loaders import (
    DirectoryLoader,
    Docx2txtLoader,
    SeleniumURLLoader,
    PyPDFLoader
)
from dotenv import find_dotenv, load_dotenv
from langchain_openai import OpenAIEmbeddings
from langchain_community.embeddings import HuggingFaceEmbeddings
from langchain.embeddings.bedrock import BedrockEmbeddings
from langchain_community.embeddings import CohereEmbeddings
from langchain_community.vectorstores import OpenSearchVectorSearch
from langchain.text_splitter import RecursiveCharacterTextSplitter
import boto3
from requests_aws4auth import AWS4Auth
from opensearchpy import RequestsHttpConnection
import requests
import xml.etree.ElementTree as ET
from aws_rag_bot.kustomer import get_all_shortcuts_as_documents
import re

DEFAULT_CHUNK_SIZE = 256


# Defined structures for the supported embedding types.  Can be extended
class EmbeddingTypes:
    OPENAI_GPT_DEFAULT = {"name": "openai-gpt-default", "provider": "openai", "model": "default"}
    HUGGING_FACE_DEFAULT = {"name": "hugging-face-default", "provider": "hugging-face", "model": "default"}
    BEDROCK_DEFAULT = {"name": "bedrock-default", "provider": "bedrock", "model": "amazon.titan-embed-text-v1",
                       "region": "us-east-1"}
    COHERE_DEFAULT = {"name": "cohere-default", "provider": "cohere", "model": "embed-english-light-v3.0"}


# ======= Static Utility Functions =======
def clean_documents_newlines_spaces_and_tabs(docs):
    # This will remove all the redundant newlines, spaces, and tabs from the documents
    #  with a good document spitter, this may not be needed, but it does reduce character count
    #     and therefore is at a minimum and small optimization
    # Compile regular expressions for matching sequences of newlines, spaces, and tabs
    newline_re = re.compile(r'\n+')
    space_re = re.compile(r' {2,}')  # Matches two or more spaces
    tab_re = re.compile(r'\t+')  # Matches any sequence of tabs

    for doc in docs:
        # Clean the document content
        cleaned_content = newline_re.sub('\n', doc.page_content)
        cleaned_content = space_re.sub(' ', cleaned_content)
        cleaned_content = tab_re.sub(' ', cleaned_content)

        # Update the document's page_content in place
        doc.page_content = cleaned_content

    return docs


def get_opensearch_endpoint(domain_name, region=None):
    # Get the callable endpoint for the OpenSearch domain
    # Note that this can be used to make Elastic Search like calls directly
    client = None
    if region:
        client = boto3.client('es', region)
    else:
        client = boto3.client('es')

    response = client.describe_elasticsearch_domain(
        DomainName=domain_name
    )

    return response['DomainStatus']['Endpoint']


def get_embeddings_from_model(embedding_model=None):
    embeddings = None
    if embedding_model is None:
        embedding_model = EmbeddingTypes.BEDROCK_DEFAULT

    if embedding_model['provider'] == 'hugging-face':
        if embedding_model['name'] == EmbeddingTypes.HUGGING_FACE_DEFAULT['name']:
            embeddings = HuggingFaceEmbeddings()
        else:
            embeddings = HuggingFaceEmbeddings(model_name=embedding_model['model'])

    elif embedding_model['provider'] == 'bedrock':
        embeddings = BedrockEmbeddings(model_id=embedding_model['model'], region_name=embedding_model['region'])

    elif embedding_model['provider'] == 'cohere':
        embeddings = CohereEmbeddings(model=embedding_model['model'])

    elif embedding_model['provider'] == 'openai':
        load_dotenv(find_dotenv())
        if embedding_model['name'] == EmbeddingTypes.OPENAI_GPT_DEFAULT['name']:
            embeddings = OpenAIEmbeddings()
        else:
            embeddings = OpenAIEmbeddings(model=embedding_model['model'])

    else:
        raise f"Error getting embedding. Cannot handle embeddings with provider {embedding_model['provider']}"

    return embeddings


def get_urls_from_sitemap(sitemap_url):
    """Fetches and extracts all URLs from a given sitemap or sitemap index."""
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3',
        'Accept': 'application/xml;q=0.9, */*;q=0.8'
    }
    if 'library' in sitemap_url:  # Hack for the wordpress site.  may work for others, but it had problems
        response = requests.get(sitemap_url, headers=headers, timeout=60)
    else:
        response = requests.get(sitemap_url)

    response.raise_for_status()

    root = ET.fromstring(response.content)

    urls = []

    # Check if it's a sitemap index
    if root.tag.endswith('sitemapindex'):
        for sitemap in root.findall("{http://www.sitemaps.org/schemas/sitemap/0.9}sitemap"):
            loc = sitemap.find("{http://www.sitemaps.org/schemas/sitemap/0.9}loc").text
            urls.extend(get_urls_from_sitemap(loc))
    # Else assume it's a regular sitemap
    else:
        for url in root.findall("{http://www.sitemaps.org/schemas/sitemap/0.9}url"):
            loc = url.find("{http://www.sitemaps.org/schemas/sitemap/0.9}loc").text
            urls.append(loc)

    return urls


def scrape_website_pages(website_urls, headless=True):
    # Uses the Selenium engine to scrape the website pages and return Langchain documents
    # https://scrapeops.io/selenium-web-scraping-playbook/python-selenium-disable-image-loading/
    loader = SeleniumURLLoader(urls=website_urls, headless=headless,
                               arguments=["enable-features=NetworkServiceInProcess"])
    docs = loader.load()

    ready_to_use_docs = []

    for doc in docs:
        ready_to_use_docs.append(doc)
    return ready_to_use_docs


def get_documents_from_folder(content_source):
    # Read a folder of documents and return Langchain documents
    #  Uses content source for better control
    #  Handles a variety of document types
    # first make sure it is valid for me to handle
    search_expression = None
    recursive = True

    directory = content_source['location']
    if content_source['type'] == 'Microsoft Word':
        search_expression = "*.docx"
        loader_class = Docx2txtLoader

    elif content_source['type'] == 'PDF':
        # NOTE: The PDF parser will break the document up into pages, so one doc could translate into many
        search_expression = "*.pdf"
        loader_class = PyPDFLoader

    else:
        raise (f"   Error loading documents. Cannot handle loading documents of type {content_source['type']}")

    if 'recursive' in content_source:
        recursive = content_source['recursive']

    print(f"   Reading {content_source['type']} documents from {content_source['location']}")
    loader = DirectoryLoader(path=directory, glob=search_expression, loader_cls=loader_class, recursive=recursive)
    folder_docs = clean_documents_newlines_spaces_and_tabs(loader.load())

    print(f"   Documents loaded in memory.  Total count={len(folder_docs)}")

    filtered_docs = folder_docs
    if 'whitelist' in content_source:
        filtered_docs = filter_urls_with_whitelist(filtered_docs, content_source['whitelist'])

    if 'blacklist' in content_source:
        filtered_docs = filter_urls_with_blacklist(filtered_docs, content_source['blacklist'])

    print(
        f"   Found {len(folder_docs)} URLs. Filted down to {len(filtered_docs)}. Lets go screen scraping (the slow part)")

    if 'chunk_size' in content_source and content_source['chunk_size'] == 0:
        return filtered_docs
    else:
        doc_chunks = chunk_up_documents(filtered_docs,
                                        content_source['chunk_size'] if 'chunk_size' in content_source else None)
        return doc_chunks


def filter_urls_with_blacklist(urls, blacklist):
    # Separate exact matches from wildcard matches
    exact_matches = [url for url in blacklist if "*" not in url]
    wildcard_matches = [url.replace("*", ".*") for url in blacklist if "*" in url]

    # Compile regex patterns for wildcard matches
    wildcard_patterns = [re.compile(pattern) for pattern in wildcard_matches]

    # Filter URLs
    filtered_urls = []
    for url in urls:
        # Check against exact matches
        if url in exact_matches:
            continue

        # Check against wildcard patterns
        if any(pattern.match(url) for pattern in wildcard_patterns):
            continue

        # If URL does not match any blacklist entry, add it to the filtered list
        filtered_urls.append(url)

    return filtered_urls


def filter_urls_with_whitelist(urls, whitelist):
    # Separate exact matches from wildcard matches
    exact_matches = [url for url in whitelist if "*" not in url]
    wildcard_matches = [url.replace("*", ".*") for url in whitelist if "*" in url]

    # Compile regex patterns for wildcard matches
    wildcard_patterns = [re.compile(pattern) for pattern in wildcard_matches]

    # Filter URLs
    filtered_urls = []
    for url in urls:
        # Check against exact matches
        if url in exact_matches:
            filtered_urls.append(url)
            continue

        # Check against wildcard patterns
        if any(pattern.match(url) for pattern in wildcard_patterns):
            filtered_urls.append(url)
            continue

    return filtered_urls


def chunk_up_documents(documents, chunk_size):
    if chunk_size is None:
        chunk_size = DEFAULT_CHUNK_SIZE

    if chunk_size == 1:
        return documents

    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap=26,
        length_function=len,
    )
    docs_chunks = text_splitter.split_documents(documents)
    print(f"   Chunked {len(documents)} documents into {len(docs_chunks)} chunks using size={chunk_size}")

    return docs_chunks


def get_documents_from_website(content_source, headless=True):
    # use the source definition to get documents form a website and return Langchain documents
    #  supports URL blacklist and whitelist filtering as well as a fixed URL list (items)
    if content_source['type'] != 'Website':
        raise (f"   Error loading documents. Cannot handle loading documents of type {content_source['type']}")

    site_urls = filtered_urls = None
    if 'location' in content_source:
        print(f"   Getting URLs from site map {content_source['location']}")
        site_urls = get_urls_from_sitemap(content_source['location'])

        # Apply filtering if defined.  First whitelist, then blacklist
        filtered_urls = site_urls
        if 'whitelist' in content_source:
            filtered_urls = filter_urls_with_whitelist(filtered_urls, content_source['whitelist'])

        if 'blacklist' in content_source:
            filtered_urls = filter_urls_with_blacklist(filtered_urls, content_source['blacklist'])

    else:
        site_urls = filtered_urls = content_source['items']

    print(
        f"   Found {len(site_urls)} URLs. Filted down to {len(filtered_urls)}. Lets go screen scraping (the slow part)")

    website_documents = clean_documents_newlines_spaces_and_tabs(scrape_website_pages(filtered_urls, headless))
    print(f"   ... and scraped pages resulting in {len(website_documents)} documents")

    if 'chunk_size' in content_source and content_source['chunk_size'] == 0:
        return website_documents
    else:
        webpage_chunks = chunk_up_documents(website_documents,
                                            content_source['chunk_size'] if 'chunk_size' in content_source else None)
        return webpage_chunks


class OpenSearchVectorDBLoader:
    __domain_name = None
    __region = None
    __index_name = None
    __embedding = None
    __data_sources = None

    def __init__(self, domain_name, index_name=None, embedding_model=None, data_sources=None, region=None):
        self.__domain_name = domain_name
        self.__region = region
        self.__index_name = index_name

        if embedding_model:
            self.__embedding_model = embedding_model
        else:
            self.__embedding_model = EmbeddingTypes.BEDROCK_DEFAULT

        self.__data_sources = data_sources

    def load_from_documents(self, documents, delete_index=False):
        load_dotenv(find_dotenv())
        embeddings = get_embeddings_from_model(self.__embedding_model)

        endpoint = get_opensearch_endpoint(self.__domain_name, self.__region)
        print(f"   Open search with endpoint={endpoint} and getting ready to load")
        print(f"   Index name = {self.__index_name} and loading {len(documents)} document chunks")

        credentials = boto3.Session().get_credentials()
        if self.__region:
            region = self.__region
        else:
            region = boto3.Session().region_name

        service = 'es'
        awsauth = AWS4Auth(credentials.access_key, credentials.secret_key, region, service,
                           session_token=credentials.token)

        # Delete the index
        if delete_index:
            self.delete_index(self.__index_name)

        batch_size = 1000
        # Calculate the total number of batches needed
        total_batches = len(documents) // batch_size + (1 if len(documents) % batch_size > 0 else 0)
        print(f"   Now loading {len(documents)} document chunks in {total_batches} batches of {batch_size}")
        for batch_num in range(total_batches):
            start_index = batch_num * batch_size
            end_index = start_index + batch_size
            document_subset = documents[start_index:end_index]

            db = OpenSearchVectorSearch.from_documents(
                document_subset,
                embeddings,  # Use the full embeddings list for each batch
                opensearch_url=f'https://{endpoint}',
                http_auth=awsauth,
                index_name=self.__index_name,
                connection_class=RequestsHttpConnection,
                verify_certs=True,
                bulk_size=batch_size
            )

            print(f"     Loaded batch {batch_num + 1} of {total_batches}")

        print(f"   Documents loaded into the DB")

        return db

    def load(self, content_sources=None, delete_index=True):
        # Get documents from each source and loads it into a vector DB
        #  Allow overriding the source and destination (index)
        #  BY default this will assume the index should be deleted and re-created.  This can be overridden
        #     by setting delete_index to False allowing multiple loads into the same index

        content_to_load = None
        if content_sources:
            content_to_load = content_sources
        else:
            if self.__data_sources:
                content_to_load = self.__data_sources
            else:
                raise (f"   Error loading documents. No content sources defined in constructor or method call")

        content_docs = None
        # delete_index = True  # in first item in loop, we clean, then we retain the index
        print(
            f"Going to load {len(self.__data_sources)} data sources into {self.__domain_name} domain and {self.__index_name} index")
        for content_source in content_to_load:
            print(f"Processing content for {content_source['name']} ")
            if content_source['type'] == 'Website':
                content_docs = get_documents_from_website(content_source)
            elif content_source['type'] in ['Microsoft Word', 'PDF']:
                content_docs = get_documents_from_folder(content_source)
            elif content_source['type'] == 'KustomerAPI':
                content_docs = get_all_shortcuts_as_documents(content_source['location'])
            else:
                raise (f"   Error loading documents. Cannot handle loading documents of type {content_source['type']}")

            print(f"Loading {len(content_docs)} content chunks from {content_source['name']} into DB")
            self.load_from_documents(content_docs, delete_index=delete_index)
            delete_index = False  # Now keep existing index data for each subsequent document set

    def delete_index(self, index_name):
        # This will delete the index if is exists.  If not, it will ignore the error
        #  requiring explicit definition of index, and not just class instance version - to avoid accidents
        credentials = boto3.Session().get_credentials()
        region = boto3.Session().region_name  # TODO: Handle this better - with region - leave as override or get from system
        service = 'es'
        awsauth = AWS4Auth(credentials.access_key, credentials.secret_key, region, service,
                           session_token=credentials.token)

        endpoint = get_opensearch_endpoint(self.__domain_name, self.__region)

        # Delete the index
        print(f"   Deleting index={index_name}")
        delete_response = requests.request("DELETE", f'https://{endpoint}/{index_name}', auth=awsauth)
        if delete_response.status_code != 200:
            print(
                f"   Error deleting index={index_name}.  Likely not found and we can ignore. Status code={delete_response.status_code} Reason={delete_response.reason}")
        else:
            print(f"   Index deleted.")

    def get_index_name(self):
        return self.__index_name


class OpenSearchVectorDBQuery:
    __domain_name = None
    __region = None
    __index_name = None
    __embedding = None

    def __init__(self, domain_name, index_name, embedding_model=None, region=None):
        self.__domain_name = domain_name
        self.__region = region
        self.__index_name = index_name
        self.__embedding_model = embedding_model
        self.__client = None

    def get_client(self):
        if self.__client:
            return self.__client

        load_dotenv(find_dotenv())
        embeddings = get_embeddings_from_model(self.__embedding_model)
        endpoint = get_opensearch_endpoint(domain_name=self.__domain_name, region=self.__region)

        credentials = boto3.Session().get_credentials()
        region = boto3.Session().region_name
        service = 'es'
        awsauth = AWS4Auth(credentials.access_key, credentials.secret_key, region, service,
                           session_token=credentials.token)

        self.__client = OpenSearchVectorSearch(
            embedding_function=embeddings,
            index_name=self.__index_name,
            opensearch_url=f'https://{endpoint}',
            http_auth=awsauth,
            use_ssl=True,
            connection_class=RequestsHttpConnection,
            verify_certs=True,
            timeout=30
        )

        return self.__client

    def query(self, query, result_set=5):
        vdb = self.get_client()
        query_results = vdb.similarity_search(query, k=result_set)

        return query_results

    def get_index_name(self):
        return self.__client.get_index_name()
